/*
 * Copyright (c) 2019 Tilman Sauerbeck (tilman at code-monkey de)
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#![no_std]
#![no_main]
#[link(name="libcommon.rlib")]

extern crate common;

mod flash;
mod bootloader;

use common::register;
use common::clock;
use common::systick;
use common::port;
use common::gpio;
use common::watchdog;
use common::usb_serial;

type Reg32 = register::Register<u32>;

extern {
    fn enable_interrupts();
    fn disable_interrupts();

    fn jump_to_application(address: u32) -> !;
}

#[cfg(bootloader_type = "intermediate")]
const APPLICATION_ADDR: u32 = 0x0;

#[cfg(bootloader_type = "final")]
const APPLICATION_ADDR: u32 = 0x8000;

#[cfg(bootloader_type = "intermediate")]
const USB_PID: u16 = 0x635b;

#[cfg(bootloader_type = "final")]
const USB_PID: u16 = 0x635c;

unsafe fn application_missing() -> bool {
    let first_app_word = Reg32::new(APPLICATION_ADDR);

    first_app_word.read() == 0xffffffff
}

#[cfg(bootloader_type = "intermediate")]
fn bootloader_requested() -> bool {
    true
}

#[cfg(bootloader_type = "final")]
fn bootloader_requested() -> bool {
    let start_ticks = systick::now();

    while !systick::has_timeout_ms(start_ticks, 3000) {
        if (gpio::get(gpio::GPIOE) & (1 << 24)) == 0 {
            return true;
        }
    }

    false
}

#[inline(never)]
#[panic_handler]
fn panic(_info: &core::panic::PanicInfo) -> ! {
    loop {
    }
}

#[no_mangle]
pub unsafe extern "C" fn _start() -> ! {
    watchdog::disable();
    clock::configure();
    systick::init();
    port::init();

    // Configure lower left push button.
    gpio::set_direction(gpio::GPIOE, 1 << 24, gpio::Direction::Input);
    port::set_af(port::PORTE, 24, 1);
    port::set_pull(port::PORTE, 24, port::Pull::Up);

    enable_interrupts();

    if application_missing() || bootloader_requested() {
        usb_serial::init(0xf055, USB_PID);

        let mut b = bootloader::Bootloader::new();

        while b.run() || application_missing() {
        }

        usb_serial::shutdown();
    }

    disable_interrupts();

    clock::reset();

    jump_to_application(APPLICATION_ADDR);
}
